using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

using Zivid.NET.HandEye;

class Program
{
    static void Main()
    {
        try
        {
            var zivid = new Zivid.NET.Application();

            Console.WriteLine("Connecting to camera...");
            var camera = zivid.ConnectCamera();
            var inputs = ReadInputs(camera);

            Console.WriteLine("Performing hand-eye calibration ... ");
            var calibrationResult = Calibrator.CalibrateEyeToHand(inputs);

            if (calibrationResult)
            {
                Console.WriteLine("{0}\n{1}\n{2}", "OK", "Result:", calibrationResult);
            }
            else
            {
                Console.WriteLine("Failed");
                Environment.ExitCode = 1;
            }
        }
        catch (Exception ex)
        {
            Console.WriteLine("Error: {0}", ex.Message);
            Environment.ExitCode = 1;
        }
    }

    static List<CalibrationInput> ReadInputs(Zivid.NET.Camera camera)
    {
        var input = new List<CalibrationInput>();
        var currentPoseId = 0U;
        var beingInput = true;

        Interaction.ExtendInputBuffer(2048);

        do
        {
            switch (Interaction.EnterCommand())
            {
                case CommandType.AddPose:
                    try
                    {
                        var robotPose = Interaction.EnterRobotPose(currentPoseId);
                        var frame = Interaction.AcquireCheckerboardFrame(camera);

                        Console.Write("Detecting checkerboard square centers... ");
                        var result = Detector.DetectFeaturePoints(frame.GetPointCloud());

                        if (result)
                        {
                            Console.WriteLine("OK");
                            input.Add(new CalibrationInput(robotPose, result));
                            ++currentPoseId;
                        }
                        else
                        {
                            Console.WriteLine("FAILED");
                        }
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Error: {0}", ex.Message);
                        continue;
                    }
                    break;

                case CommandType.Calibrate:
                    beingInput = false;
                    break;

                case CommandType.Unknown:
                    Console.WriteLine("Error: Unknown command");
                    break;
            }
        } while (beingInput);
        return input;
    }
}

enum CommandType
{
    AddPose,
    Calibrate,
    Unknown
}

class Interaction
{
    // Console.ReadLine only supports reading 256 characters, by default. This limit is modified
    // by calling ExtendInputBuffer with the maximum length of characters to be read.
    public static void ExtendInputBuffer(int size)
    {
        Console.SetIn(new StreamReader(Console.OpenStandardInput(), Console.InputEncoding, false, size));
    }

    public static CommandType EnterCommand()
    {
        Console.Write("Enter command, p (to add robot pose) or c (to perform calibration): ");
        var command = Console.ReadLine().ToLower();

        switch (command)
        {
            case "p": return CommandType.AddPose;
            case "c": return CommandType.Calibrate;
            default: return CommandType.Unknown;
        }
    }

    public static Pose EnterRobotPose(ulong index)
    {
        var elementCount = 16;
        Console.WriteLine("Enter pose with id (a line with {0} space separated values describing 4x4 row-major matrix) : {1}", elementCount, index);
        var input = Console.ReadLine();

        var elements = input.Split()
            .Where(x => !string.IsNullOrEmpty(x.Trim()))
            .Select(x => double.Parse(x))
            .ToArray();

        var robotPose = new Pose(elements);
        Console.WriteLine("The following pose was entered: \n{0}", robotPose);
        return robotPose;
    }

    public static Zivid.NET.Frame AcquireCheckerboardFrame(Zivid.NET.Camera camera)
    {
        Console.Write("Capturing checkerboard image... ");
        var settings = new Zivid.NET.Settings
        {
            Iris = 17,
            Gain = 1.0,
            Brightness = 1.0,
            ExposureTime = Zivid.NET.Duration.FromMicroseconds(20_000),
        };
        settings.Filters.Gaussian.Enabled = true;

        camera.SetSettings(settings);
        var frame = camera.Capture();

        Console.WriteLine("OK");
        return frame;
    }
}
